from homeconnect_webthing.abstract_app import AbstractApp, ArgumentSpec
from homeconnect_webthing.homeappliances_webthing import run_server
from typing import Dict, Any

class MyApp(AbstractApp):

    def do_listen(self, args: Dict[str, Any]) -> bool:
        print('starting webthing server on port ' + str(args['port']))
        run_server(args['port'], description=self.description)
        return True


def main():
    MyApp("homeconnect_webthing", [ArgumentSpec("period", int, "the period", 56, True)]).handle_command()


if __name__ == '__main__':
    main()



from os import system, remove
from os import listdir
from abc import ABC, abstractmethod
import pathlib
import logging
import subprocess
import argparse
from dataclasses import dataclass
from importlib.metadata import metadata, entry_points
from typing import List, Any, Dict



class Unit:

    UNIT_TEMPLATE = '''
    [Unit]
    Description=$packagename
    After=syslog.target
    
    [Service]
    Type=simple
    ExecStart=$entrypoint --command listen $params
    SyslogIdentifier=$packagename
    StandardOutput=syslog
    StandardError=syslog
    Restart=always
    RestartSec=3
    
    [Install]
    WantedBy=multi-user.target
    '''


    def __init__(self, packagename: str):
        self.packagename = packagename

    def __print_status(self, service: str):
        try:
            status = subprocess.check_output("sudo systemctl is-active " + service, shell=True, stderr=subprocess.STDOUT)
            if status.decode('ascii').strip() == 'active':
                print(service + " is running (print log by calling " + "sudo journalctl -n 20 -u " + service + ")")
                return
        except subprocess.CalledProcessError as e:
            pass
        print("Warning: " + service + " is not running")
        system("sudo journalctl -n 20 -u " + service)

    def register(self, entrypoint: str, args: Dict[str, Any]):
        service = self.servicename(args['port'])
        replacements = {'packagename': self.packagename, 'entrypoint': entrypoint}
        unit = Unit.UNIT_TEMPLATE
        params = []
        for name in args.keys():
            if name != "command":
                params.append("--" + name + " " + str(args[name]))
        replacements['params'] = " ".join(params)
        for name in replacements.keys():
            unit = unit.replace("$" + name, replacements[name])
        unit_file_fullname = str(pathlib.Path("/", "etc", "systemd", "system", service))
        with open(unit_file_fullname, "w") as file:
            file.write(unit)
        system("sudo systemctl daemon-reload")
        system("sudo systemctl enable " + service)
        system("sudo systemctl restart " + service)
        self.__print_status(service)

    def deregister(self, args: Dict[str, Any]):
        service = self.servicename(args['port'])
        unit_file_fullname = str(pathlib.Path("/", "etc", "systemd", "system", service))
        system("sudo systemctl stop " + service)
        system("sudo systemctl disable " + service)
        system("sudo systemctl daemon-reload")
        try:
            remove(unit_file_fullname)
        except Exception as e:
            pass

    def printlog(self, args: Dict[str, Any]):
        service = self.servicename(args['port'])
        system("sudo journalctl -f -u " + service)

    def servicename(self, port: int):
        return self.packagename + "_" + str(port) + ".service"

    def list_installed(self):
        services = []
        try:
            for file in listdir(pathlib.Path("/", "etc", "systemd", "system")):
                if file.startswith(self.packagename) and file.endswith('.service'):
                    idx = file.rindex('_')
                    port = str(file[idx+1:file.index('.service')])
                    services.append((file, port, self.is_active(file)))
        except Exception as e:
            pass
        return services

    def is_active(self, servicename: str):
        cmd = '/bin/systemctl status %s' % servicename
        proc = subprocess.Popen(cmd, shell=True,stdout=subprocess.PIPE,encoding='utf8')
        stdout_list = proc.communicate()[0].split('\n')
        for line in stdout_list:
            if 'Active:' in line:
                if '(running)' in line:
                    return True
        return False


@dataclass
class ArgumentSpec:
    name: str
    dt: type
    description: str
    default_value: Any = None
    required: bool = False

    def resolve(self, args):
        return vars(args)[self.name]


class AbstractApp(ABC):

    def __init__(self, packagename: str, arg_specs: List[ArgumentSpec] = list(), default_port: int = 8644):
        self.unit = Unit(packagename)
        self.packagename = packagename
        self.arg_specs = arg_specs
        self.default_port = default_port
        md = metadata(packagename)
        self.description = md.get('description', "")
        for script in entry_points()['console_scripts']:
            if script.value == packagename + '.__main__:main':
                self.entrypoint = script.name
        print(self.description)

    def parse_arguments(self) -> Dict[str, Any]:
        parser = argparse.ArgumentParser(description=self.description)
        parser.add_argument('--command', metavar='command', required=False, type=str, help='the command. Supported commands are: listen (run the webthing service), register (register and starts the webthing service as a systemd unit, deregister (deregisters the systemd unit), log (prints the log)')
        parser.add_argument('--port', metavar='port', required=False, type=int, default=self.default_port, help='the port of the webthing serivce')
        parser.add_argument('--verbose', metavar='verbose', required=False, type=bool, default=False, help='activates verbose output')
        for spec in self.arg_specs:
            parser.add_argument('--' + spec.name, metavar=spec.name, required=False, type=spec.dt, default=spec.default_value, help=spec.description)
        args = parser.parse_args()
        arguments = {"port": args.port, "verbose": args.verbose, "command": args.command}
        for arg_spec in self.arg_specs:
            arguments[arg_spec.name] = arg_spec.resolve(args)
        return arguments

    def handle_command(self):
        args = self.parse_arguments()
        if args['verbose']:
            log_level=logging.DEBUG
        else:
            log_level=logging.INFO
        logging.basicConfig(format='%(asctime)s %(name)-20s: %(levelname)-8s %(message)s', level=log_level, datefmt='%Y-%m-%d %H:%M:%S')

        handled = False
        if args['command'] == 'listen':
            handled = self.do_listen(args)
        elif args['command'] == 'register':
            handled = self.do_register(args)
        elif args['command'] == 'deregister':
            handled = self.do_register(args)
        if not handled:
            self.do_print_usage_info(args)

    def do_print_usage_info(self, args: Dict[str, Any]) -> bool:
        print("for command options usage")
        print(" sudo " + self.entrypoint + " --help")
        print("example commands")
        print(" sudo " + self.entrypoint + " --command register --port " + str(args['port']) + " " + " ".join(["--" + argument.name + " " + str(argument.default_value) for argument in self.arg_specs]))
        print(" sudo " + self.entrypoint + " --command listen --port " + str(args['port']) + " " +  " ".join(["--" + argument.name + " " + str(argument.default_value) for argument in self.arg_specs]))
        if len(self.unit.list_installed()) > 0:
            print("example commands for registered services")
            for service_info in self.unit.list_installed():
                port = service_info[1]
                print(" sudo " + self.entrypoint + " --command deregister --port " + port)
                print(" sudo " + self.entrypoint + " --command log --port " + port)
        return True

    def do_listen(self, args: Dict[str, Any]) -> bool:
        return False

    def do_register(self, args: Dict[str, Any]) -> bool:
        print('register and starting webthing server on port ' + str(args['port']))
        Unit(self.packagename).register(self.entrypoint, args)
        return True

    def do_deregister(self, args: Dict[str, Any]) -> bool:
        print('deregister ' + str(args['port']))
        Unit(self.packagename).deregister(args)
        return True




from webthing import (MultipleThings, Property, Thing, Value, WebThingServer)
import logging
import tornado.ioloop
from datetime import datetime
from homeconnect_webthing.homeappliances import HomeConnect, Dishwasher



class DishwasherThing(Thing):

    # regarding capabilities refer https://iot.mozilla.org/schemas
    # there is also another schema registry http://iotschema.org/docs/full.html not used by webthing

    def __init__(self, description: str, dishwasher: Dishwasher):
        Thing.__init__(
            self,
            'urn:dev:ops:Dishwasher-1',
            'Dishwasher',
            ['MultiLevelSensor'],
            description
        )

        self.ioloop = tornado.ioloop.IOLoop.current()

        self.dishwasher = dishwasher
        dishwasher.register_value_changed_listener(self.on_value_changed)


        self.name = Value(dishwasher.name)
        self.add_property(
            Property(self,
                     'device_name',
                     self.name,
                     metadata={
                         'title': 'Name',
                         "type": "string",
                         'description': 'The device name',
                         'readOnly': True,
                     }))

        self.device_type = Value(dishwasher.device_type)
        self.add_property(
            Property(self,
                     'device_type',
                     self.device_type,
                     metadata={
                         'title': 'Type',
                         "type": "string",
                         'description': 'The device type',
                         'readOnly': True,
                     }))


        self.haid = Value(dishwasher.haid)
        self.add_property(
            Property(self,
                     'device_haid',
                     self.haid,
                     metadata={
                         'title': 'haid',
                         "type": "string",
                         'description': 'The device haid',
                         'readOnly': True,
                     }))

        self.brand = Value(dishwasher.brand)
        self.add_property(
            Property(self,
                     'device_brand',
                     self.brand,
                     metadata={
                         'title': 'Brand',
                         "type": "string",
                         'description': 'The device brand',
                         'readOnly': True,
                     }))

        self.vib = Value(dishwasher.vib)
        self.add_property(
            Property(self,
                     'device_vib',
                     self.vib,
                     metadata={
                         'title': 'Vib',
                         "type": "string",
                         'description': 'The device vib',
                         'readOnly': True,
                     }))

        self.enumber = Value(dishwasher.enumber)
        self.add_property(
            Property(self,
                     'device_enumber',
                     self.enumber,
                     metadata={
                         'title': 'Enumber',
                         "type": "string",
                         'description': 'The device enumber',
                         'readOnly': True,
                     }))

        self.power = Value(dishwasher.power)
        self.add_property(
            Property(self,
                     'power',
                     self.power,
                     metadata={
                         'title': 'Power State',
                         "type": "string",
                         'description': 'The power state. See https://api-docs.home-connect.com/settings?#power-state',
                         'readOnly': True,
                     }))

        self.door = Value(dishwasher.door)
        self.add_property(
            Property(self,
                     'door',
                     self.door,
                     metadata={
                         'title': 'Door State',
                         "type": "string",
                         'description': 'Door State. See https://api-docs.home-connect.com/states?#door-state',
                         'readOnly': True,
                     }))

        self.operation = Value(dishwasher.operation)
        self.add_property(
            Property(self,
                     'operation',
                     self.operation,
                     metadata={
                         'title': 'Operation State',
                         "type": "string",
                         'description': 'The operation state. See https://api-docs.home-connect.com/states?#operation-state',
                         'readOnly': True,
                     }))

        self.remote_start_allowed = Value(dishwasher.remote_start_allowed)
        self.add_property(
            Property(self,
                     'remote_start_allowed',
                     self.remote_start_allowed,
                     metadata={
                         'title': 'Remote Start Allowed State',
                         "type": "boolean",
                         'description': 'Remote Start Allowance State. See https://api-docs.home-connect.com/states?#remote-start-allowance-state',
                         'readOnly': True,
                     }))

        self.selected_program = Value(dishwasher.program_selected)
        self.add_property(
            Property(self,
                     'program_selected',
                     self.selected_program,
                     metadata={
                         'title': 'Selected Program',
                         "type": "string",
                         'description': 'Selected Program',
                         'readOnly': True,
                     }))

        self.program_vario_speed_plus = Value(dishwasher.program_vario_speed_plus)
        self.add_property(
            Property(self,
                     'program_vario_speed_plus',
                     self.program_vario_speed_plus,
                     metadata={
                         'title': 'program_vario_speed_plus',
                         "type": "boolean",
                         'description': 'VarioSpeed Plus Option. See https://api-docs.home-connect.com/programs-and-options?#dishwasher_variospeed-plus-option',
                         'readOnly': True,
                     }))


        self.program_hygiene_plus = Value(dishwasher.program_hygiene_plus)
        self.add_property(
            Property(self,
                     'program_hygiene_plus',
                     self.program_hygiene_plus,
                     metadata={
                         'title': 'program_hygiene_plus',
                         "type": "boolean",
                         'description': 'Hygiene Plus Option',
                         'readOnly': True,
                     }))


        self.program_extra_try = Value(dishwasher.program_extra_try)
        self.add_property(
            Property(self,
                     'program_extra_try',
                     self.program_extra_try,
                     metadata={
                         'title': 'program_extra_try',
                         "type": "boolean",
                         'description': 'Extra Try Option',
                         'readOnly': True,
                     }))

        self.start_date = Value(dishwasher.start_date, dishwasher.start_date)
        self.add_property(
            Property(self,
                     'program_start_date',
                     self.start_date,
                     metadata={
                         'title': 'Start date',
                         "type": "string",
                         'description': 'Th start date',
                         'readOnly': True,
                     }))

        self.program_progress = Value(dishwasher.program_progress)
        self.add_property(
            Property(self,
                     'program_progress',
                     self.program_progress,
                     metadata={
                         'title': 'Progress',
                         "type": "number",
                         'description': 'progress',
                         'readOnly': True,
                     }))

    def on_value_changed(self):
        self.ioloop.add_callback(self.__on_value_changed)

    def __on_value_changed(self):
        self.power.notify_of_external_update(self.dishwasher.power)
        self.door.notify_of_external_update(self.dishwasher.door)
        self.operation.notify_of_external_update(self.dishwasher.operation)
        self.remote_start_allowed.notify_of_external_update(self.dishwasher.remote_start_allowed)
        self.enumber.notify_of_external_update(self.dishwasher.enumber)
        self.selected_program.notify_of_external_update(self.dishwasher.program_selected)
        self.program_vario_speed_plus.notify_of_external_update(self.dishwasher.program_vario_speed_plus)
        self.program_hygiene_plus.notify_of_external_update(self.dishwasher.program_hygiene_plus)
        self.program_extra_try.notify_of_external_update(self.dishwasher.program_extra_try)
        self.program_progress.notify_of_external_update(self.dishwasher.program_progress)
        self.vib.notify_of_external_update(self.dishwasher.vib)
        self.brand.notify_of_external_update(self.dishwasher.brand)
        self.haid.notify_of_external_update(self.dishwasher.haid)
        self.name.notify_of_external_update(self.dishwasher.name)
        self.device_type.notify_of_external_update(self.dishwasher.device_type)
        self.start_date.notify_of_external_update(self.dishwasher.start_date)


def run_server(port: int, description: str):
    homeappliances = []
    for device in HomeConnect().devices():
        if device.is_dishwasher():
            homeappliances.append(DishwasherThing(description, device))
    homeappliances.sort()
    logging.info(str(len(homeappliances)) + " homeappliances found: " + ", ".join([homeappliance.dishwasher.name + "/" + homeappliance.dishwasher.enumber for homeappliance in homeappliances]))
    server = WebThingServer(MultipleThings(homeappliances, 'homeappliances'), port=port, disable_host_validation=True)
    logging.info('running webthing server http://localhost:' + str(port))
    try:
        server.start()
    except KeyboardInterrupt:
        logging.info('stopping webthing server')
        server.stop()
        logging.info('done')



import logging
import requests
import webbrowser
from os import path
from datetime import datetime, timedelta
from urllib.parse import urlsplit, parse_qs
from typing import Optional, Tuple



class Auth:

    URI = "https://api.home-connect.com/security"
    DEFAULT_FILENAME = "homeconnect_oauth.txt"

    @staticmethod
    def create(client_id: str, client_secret:str, scope: str = "IdentifyAppliance%20Dishwasher"):
        uri = Auth.URI + "/oauth/authorize?response_type=code&client_id=" + client_id + "&scope=" + scope
        webbrowser.open(uri)

        auth_result = input("Please enter the URL redirected to: ")
        query = urlsplit(auth_result).query
        params = parse_qs(query)
        authorization_code = params['code']

        data = {"client_id": client_id,
                "client_secret": client_secret,
                "grant_type": "authorization_code",
                "code": authorization_code}
        response = requests.post(Auth.URI + '/oauth/token', data=data)
        data = response.json()
        return Auth(data['refresh_token'], client_secret)

    def __init__(self, refresh_token: str, client_secret: str):
        self.__refresh_token = refresh_token
        self.__client_secret = client_secret
        self.__access_token, self.__expiring_date = self.__refresh_access_token()

    @property
    def access_token(self) -> str:
        if datetime.now() > (self.__expiring_date + timedelta(seconds=60)):
            self.__refresh_access_token()
        return self.__access_token

    def __refresh_access_token(self) -> Tuple[str, datetime]:
        data = {"grant_type": "refresh_token",
                "refresh_token": self.__refresh_token,
                "client_secret": self.__client_secret}
        response = requests.post(Auth.URI + '/oauth/token', data=data)
        response.raise_for_status()
        data = response.json()
        expiring_date = datetime.now() + timedelta(seconds=data['expires_in'])
        return data['access_token'], expiring_date

    def store(self, filename : str = DEFAULT_FILENAME):
        logging.info("storing secret file " + path.abspath(filename))
        with open(filename, "w") as file:
            file.write("refresh_token: " + self.__refresh_token + "\n")
            file.write("client_secret: " + self.__client_secret + "\n")

    @staticmethod
    def load(filename : str = DEFAULT_FILENAME) -> Optional:
        if path.isfile(filename):
            logging.info("loading secret file " + path.abspath(filename))
            with open(filename, "r") as file:
                refresh_line = file.readline()
                refresh_token = refresh_line[refresh_line.index(":")+1:].strip()
                client_secret_line = file.readline()
                client_secret = client_secret_line[client_secret_line.index(":")+1:].strip()
                return Auth(refresh_token, client_secret)
        else:
            logging.info("secret file " + path.abspath(filename) + " does not exist")
            return None

    def __str__(self):
        return "refresh_token: " + self.__refresh_token + "\n" + "client_secret: " + self.__client_secret

    def __repr__(self):
        return self.__str__()



import logging
import requests
import sseclient
import json
from time import sleep
from abc import ABC
from threading import Thread
from typing import List, Optional, Dict, Any
from datetime import datetime, timedelta
from homeconnect_webthing.auth import Auth




class EventListener(ABC):

    def on_keep_alive_event(self, event):
        pass

    def on_notify_event(self, event):
        pass

    def on_status_event(self, event):
        pass


class Device(EventListener):

    def __init__(self, uri: str, auth: Auth, name: str, device_type: str, haid: str, brand: str, vib: str, enumber: str):
        self._uri = uri
        self._auth = auth
        self.name = name
        self.device_type = device_type
        self.haid = haid
        self.brand = brand
        self.vib = vib
        self.enumber = enumber

    def is_dishwasher(self) -> bool:
        return False

    def _perform_get(self, path:str) -> Dict[str, Any]:
        uri = self._uri + path
        logging.info("query GET " + uri)
        response = requests.get(uri, headers={"Authorization": "Bearer " + self._auth.access_token})
        response.raise_for_status()
        data = response.json()
        return data

    @property
    def __fingerprint(self) -> str:
        return self.device_type + ":" + self.brand + ":" + self.vib + ":" + self.enumber + ":" + self.haid

    def __hash__(self):
        return hash(self.__fingerprint)

    def __lt__(self, other):
        return self.__fingerprint < other.__fingerprint

    def __eq__(self, other):
        return self.__fingerprint == other.__fingerprint

    def __str__(self):
            return self.name + " (" + self.haid + ")"

    def __repr__(self):
        return self.__str__()



class Dishwasher(Device):

    def __init__(self, uri: str, auth: Auth, name: str, device_type: str, haid: str, brand: str, vib: str, enumber: str):
        self.date_refreshed = datetime.now()
        self.__power = ""
        self.__operation = ""
        self.__door = ""
        self.__program_selected = ""
        self.remote_start_allowed = False
        self.__program_start_in_relative_sec = ""
        self.__program_remaining_time = ""
        self.program_progress = None
        self.__program_remote_control_active = ""
        self.program_extra_try = ""
        self.program_hygiene_plus = ""
        self.program_vario_speed_plus = ""
        self._value_changed_listeners = set()
        super().__init__(uri, auth, name, device_type, haid, brand, vib, enumber)
        self.__refresh()

    def is_dishwasher(self) -> bool:
        return True

    def register_value_changed_listener(self, value_changed_listener):
        self._value_changed_listeners.add(value_changed_listener)

    def on_notify_event(self, event):
        self.on__value_changed_event(event)

    def on_status_event(self, event):
        self.on__value_changed_event(event)

    def on__value_changed_event(self, event):
        if event.id == self.haid:
            try:
                logging.info("event received: " + str(event))
                logging.info("data: " + str(event.data))
                data = json.loads(event.data)
                self.__on_value_changes(data.get('items', []))
                for value_changed_listener in self._value_changed_listeners:
                    value_changed_listener()
            except Exception as e:
                logging.warning("error occurred by handling event " + str(event), e)

    def __on_value_changes(self, changes: List[Any]):
        for record in changes:
            if record['key'] == 'BSH.Common.Status.DoorState':
                self.__door = record['value']
            elif record['key'] == 'BSH.Common.Status.OperationState':
                self.__operation = record['value']
            elif record['key'] == 'BSH.Common.Status.RemoteControlStartAllowed':
                self.remote_start_allowed = record['value']
            elif record['key'] == 'BSH.Common.Setting.PowerState':
                self.__power = record['value']
            elif record['key'] == 'BSH.Common.Option.StartInRelative':
                self.__program_start_in_relative_sec = record['value']
            elif record['key'] == 'BSH.Common.Option.RemainingProgramTime':
                self.__program_remaining_time = record['value']
            elif record['key'] == 'BSH.Common.Option.ProgramProgress':
                self.program_progress = record['value']
            elif record['key'] == 'BSH.Common.Status.RemoteControlActive':
                self.__program_remote_control_active = record['value']
            elif record['key'] == 'Dishcare.Dishwasher.Option.ExtraDry':
                self.program_extra_try = record['value']
            elif record['key'] == 'Dishcare.Dishwasher.Option.HygienePlus':
                self.program_hygiene_plus = record['value']
            elif record['key'] == 'Dishcare.Dishwasher.Option.VarioSpeedPlus':
                self.program_vario_speed_plus = record['value']
            else:
                print(record)

    def on_keep_alive_event(self, event):
        if datetime.now() > (self.date_refreshed + timedelta(minutes=45)):
            logging.info("refresh state")
            self.__refresh()

    def __refresh(self):
        settings = self._perform_get('/settings')['data']['settings']
        self.__on_value_changes(settings)

        status = self._perform_get('/status')['data']['status']
        self.__on_value_changes(status)

        record = self._perform_get('/programs/selected')['data']
        self.__program_selected = record['key']
        self.__on_value_changes(record['options'])

        self.date_refreshed = datetime.now()

    @property
    def power(self):
        return self.__power[self.__power.rindex('.')+1:]

    @property
    def door(self):
        return self.__door[self.__door.rindex('.')+1:]

    @property
    def operation(self):
        return self.__operation[self.__operation.rindex('.')+1:]

    @property
    def program_selected(self):
        return self.__program_selected[self.__program_selected.rindex('.')+1:]

    @property
    def start_date(self) -> str:
        start_date = (datetime.now() + timedelta(seconds=self.__program_start_in_relative_sec))
        if start_date > datetime.now():
            return start_date.strftime("%Y-%m-%dT%H:%M")
        else:
            return None

    @start_date.setter
    def start_date(self, dt: str):
        if self.operation == "BSH.Common.EnumType.OperationState.Run":
            logging.info("dishwasher is already running")
        else:
            remaining_secs_to_wait = (datetime.fromisoformat(dt) - datetime.now()).seconds
            uri = self._uri + "/programs/active"
            data = {
                "data": {
                    "key": self.program_selected,
                    "options": [ {
                                    "key": "BSH.Common.Option.StartInRelative",
                                    "value": str(remaining_secs_to_wait),
                                    "unit": "seconds"
                                } ]
                }
            }
            logging.info("query PUT " + uri)
            js = json.dumps(data)
            response = requests.put(uri, data=js, headers={"Content-Type": "application/json", "Authorization": "Bearer " + self._auth.access_token})
            response.raise_for_status()
            self.__refresh()

    def __str__(self):
        return "power=" + str(self.power) + \
               "operation=" + str(self.operation) + \
               "\ndoor=" + str(self.door) + \
               "\nremote_start_allowed=" + str(self.remote_start_allowed) + \
               "\nprogram_selected=" + str(self.program_selected)

    def __repr__(self):
        return self.__str__()


def create_device(uri: str, auth: Auth, name: str, device_type: str, haid: str, brand: str, vib: str, enumber: str) -> Device:
    if device_type.lower() == "dishwasher":
        return Dishwasher(uri, auth, name, device_type, haid, brand, vib, enumber)
    else:
        return Device(uri, auth, name, device_type, haid, brand, vib, enumber)


class HomeConnect:

    API_URI = "https://api.home-connect.com/api"

    def __init__(self):
        self.notify_listeners: List[EventListener] = list()
        self.auth = Auth.load()
        if self.auth == None:
            refresh_token = input("Please enter refresh token: ").strip()
            client_secret = input("Please enter client secret: ").strip()
            self.auth = Auth(refresh_token, client_secret)
            self.auth.store()
        Thread(target=self.__listening_for_events, daemon=True).start()

    def __listening_for_events(self):
        sleep(3)
        uri = HomeConnect.API_URI + "/homeappliances/events"
        num_reconnect = 0
        while True:
            try:
                num_reconnect += 1
                logging.info("opening sse socket to " + uri)
                response = requests.get(uri, stream=True, headers={'Accept': 'text/event-stream', "Authorization": "Bearer " + self.auth.access_token})
                response.raise_for_status()
                client = sseclient.SSEClient(response)

                num_reconnect = 0
                logging.info("consuming event...")
                for event in client.events():
                    if event.event == "NOTIFY":
                        for notify_listener in self.notify_listeners:
                            notify_listener.on_notify_event(event)
                    elif event.event == "KEEP-ALIVE":
                        for notify_listener in self.notify_listeners:
                            notify_listener.on_keep_alive_event(event)
                    elif event.event == "STATUS":
                        for notify_listener in self.notify_listeners:
                            notify_listener.on_status_event(event)
                    else:
                        print(event)
            except Exception as e:
                logging.warning("Error occurred by opening sse socket to " + uri + " " + str(e))
                wait_time_sec = {0: 3, 1:5, 2: 30, 3: 2*60, 4: 5*60}.get(num_reconnect, 30*60)
                logging.info("try reconnect in " + str(wait_time_sec) + "sec")
                sleep(wait_time_sec)

    def devices(self) -> List[Device]:
        uri = HomeConnect.API_URI + "/homeappliances"
        logging.info("requesting " + uri)
        response = requests.get(uri, headers={"Authorization": "Bearer " + self.auth.access_token})
        response.raise_for_status()
        data = response.json()
        devices = list()
        for homeappliances in data['data']['homeappliances']:
            device = create_device(HomeConnect.API_URI + "/homeappliances/" + homeappliances['haId'],
                                   self.auth,
                                   homeappliances['name'],
                                   homeappliances['type'],
                                   homeappliances['haId'],
                                   homeappliances['brand'],
                                   homeappliances['vib'],
                                   homeappliances['enumber'])
            self.notify_listeners.append(device)
            devices.append(device)
        return devices

    def dishwashers(self) -> List[Dishwasher]:
        return [device for device in self.devices() if isinstance(device, Dishwasher)]

    def dishwasher(self) -> Optional[Dishwasher]:
        dishwashers = self.dishwashers()
        if len(dishwashers) > 0:
            return dishwashers[0]
        else:
            return None

